(* version *)

let ver = Yaml.yaml_get_version_string () in
assert (ver = "0.1.3");;
let (ma, mi, pa) = Yaml.yaml_get_version () in
assert (ma = 0 && mi = 1 && pa = 3);;

(* parser *)

let read data cursor dest index length = (
	let really_length = (
		if !cursor + length <= String.length data then length else
		String.length data - !cursor
	) in
	String.blit data !cursor dest index really_length;
	cursor := !cursor + really_length;
	really_length
);;

let data = "a: b\nc: d" in
let parser = Yaml.yaml_parser_initialize (read data (ref 0)) in
let rec loop rs = (
	let event = Yaml.yaml_parser_parse parser in
	let r2 = event :: rs in
	match event with
	| Yaml.YAML_STREAM_END_EVENT, _, _ -> r2
	| _ -> loop r2
) in
let rs = List.rev (loop []) in
assert (rs = [
	Yaml.YAML_STREAM_START_EVENT
		Yaml.YAML_UTF8_ENCODING,
		(0, 0, 0), (0, 0, 0);
	Yaml.YAML_DOCUMENT_START_EVENT
		(None, [], true),
		(0, 0, 0), (0, 0, 0);
	Yaml.YAML_MAPPING_START_EVENT
		("", "", true, Yaml.YAML_BLOCK_MAPPING_STYLE),
		(0, 0, 0), (0, 0, 0);
	Yaml.YAML_SCALAR_EVENT
		("", "", "a", true, false, Yaml.YAML_PLAIN_SCALAR_STYLE),
		(0, 0, 0), (1, 0, 1);
	Yaml.YAML_SCALAR_EVENT
		("", "", "b", true, false, Yaml.YAML_PLAIN_SCALAR_STYLE),
		(3, 0, 3), (4, 0, 4);
	Yaml.YAML_SCALAR_EVENT
		("", "", "c", true, false, Yaml.YAML_PLAIN_SCALAR_STYLE),
		(5, 1, 0), (6, 1, 1);
	Yaml.YAML_SCALAR_EVENT
		("", "", "d", true, false, Yaml.YAML_PLAIN_SCALAR_STYLE),
		(8, 1, 3), (9, 1, 4);
	Yaml.YAML_MAPPING_END_EVENT,
		(9, 2, 0), (9, 2, 0);
	Yaml.YAML_DOCUMENT_END_EVENT
		true,
		(9, 2, 0), (9, 2, 0);
	Yaml.YAML_STREAM_END_EVENT,
		(9, 2, 0), (9, 2, 0)]);;

(* emitter *)

let buf = Buffer.create 1000 in
let emitter = Yaml.yaml_emitter_initialize (Buffer.add_substring buf) in
List.iter (Yaml.yaml_emitter_emit emitter) [
	Yaml.YAML_STREAM_START_EVENT Yaml.YAML_ANY_ENCODING;
	Yaml.YAML_DOCUMENT_START_EVENT (None, [], false);
	Yaml.YAML_MAPPING_START_EVENT ("", "", false, Yaml.YAML_ANY_MAPPING_STYLE);
	Yaml.YAML_ALIAS_EVENT "X";
	Yaml.YAML_SEQUENCE_START_EVENT ("", "", false, Yaml.YAML_ANY_SEQUENCE_STYLE);
	Yaml.YAML_SCALAR_EVENT ("", Yaml.yaml_str_tag, "Y", false, false, Yaml.YAML_ANY_SCALAR_STYLE);
	Yaml.YAML_SEQUENCE_END_EVENT;
	Yaml.YAML_MAPPING_END_EVENT;
	Yaml.YAML_DOCUMENT_END_EVENT false;
	Yaml.YAML_STREAM_END_EVENT];
Yaml.yaml_emitter_flush emitter;
assert (Buffer.contents buf = 
	"---\n" ^
	"*X:\n" ^
	"- !!str Y\n" ^
	"...\n");

(* serialize *)

type variant = A | B;;

let string_of_variant (x: variant): string = (
	match x with
	| A -> "A"
	| B -> "B"
);;

let variant_of_string (x: string): variant = (
	match x with
	| "A" -> A
	| "B" -> B
	| _ -> failwith "variant_of_string"
);;

let io_variant ?key v s = Serialize.make_io
	(fun x -> `tag (string_of_variant x))
	(fun x -> match x with `tag s -> variant_of_string s | _ -> failwith "io_variant")
	?key v s;;

type t = {w: variant; x: float; y: string; z: char option};;

let io_t ?key v s = Serialize.io_record (fun v s -> {
	w = io_variant ~key:"w" v.w s;
	x = Serialize.io_float ~key:"x" v.x s;
	y = Serialize.io_string ~key:"y" v.y s;
	z = Serialize.io_option Serialize.io_char '\x00' ~key:"z" v.z s}) ?key v s;;

let data1 = {w = A; x = 1.0; y = "hello, world!"; z = Some 'A'};;
let data2 = {w = B; x = -0.5; y = ""; z = None};;

let buf = Buffer.create 1024 in
let w = Serialize_yaml.Writer.create "!TEST" (Buffer.add_substring buf) in
let (_: t) = io_t data1 w in
let yaml = Buffer.contents buf in
assert (yaml =
	"--- !TEST\n" ^
	"w: A\n" ^
	"x: 1.\n" ^
	"y: \"hello, world!\"\n" ^
	"z: \'A\'\n");
let r = Serialize_yaml.Reader.create "!TEST" (read yaml (ref 0)) in
let result = io_t data2 r in
assert (result = data1);;

let buf = Buffer.create 1024 in
let w = Serialize_yaml.Writer.create "!TEST" (Buffer.add_substring buf) in
let (_: string) = Serialize.io_string (
	"智に働けば角が立つ。\n" ^
	"情に棹させば流される。\n" ^
	"意地を通せば窮屈だ。\n" ^
	"とかくに人の世は住みにくい。\n") w in
assert (Buffer.contents buf =
	"--- !TEST |\n" ^
	"  智に働けば角が立つ。\n" ^
	"  情に棹させば流される。\n" ^
	"  意地を通せば窮屈だ。\n" ^
	"  とかくに人の世は住みにくい。\n");;
