open Dfm;;

let extended_of_string: (string -> float * (int64 * int)) ref = ref
	begin fun s ->
		let x = float_of_string s in
		x, extended_of_float x
	end;;

let string_of_extended: (float * (int64 * int) -> string) ref = ref
	(* precision of Printf.printf "%.18f" is very bad... *)
	begin fun (x, _) ->
		let d = "0123456789" in
		let b = Buffer.create 32 in
		let a = (
			if x >= 0.0 then x else (
				Buffer.add_char b '-';
				~-. x
			)
		) in
		let i = floor a in
		if i = 0.0 then Buffer.add_char b '0' else (
			let string_rev s = (
				let l = String.length s in
				let r = String.create l in
				let h = l - 1 in
				for i = 0 to h do
					r.[i] <- s.[h - i]
				done;
				r
			) in
			let rec iloop b n = (
				if n = 0.0 then Buffer.contents b else (
					let f = mod_float n 10.0 in
					Buffer.add_char b d.[int_of_float f];
					iloop b ((n -. f) /. 10.0)
				)
			) in
			Buffer.add_string b (string_rev (iloop (Buffer.create 32) i))
		);
		Buffer.add_char b '.';
		let p = a -. i in
		let rec ploop b n fix i = (
			if i = 0 then () else (
				let t = n *. 10.0 in
				let f = min (floor (t +. fix)) 9.0 in
				Buffer.add_char b d.[int_of_float f];
				ploop b (max (t -. f) 0.0) (min (fix *. 100.0) 1e-2) (i - 1)
			)
		) in
		(* simulate characteristics of Delphi's formatting *)
		let mx = mod_float x 0.0001 in
		let m9 =
			9.99999999896444917e-05 ::
			9.99999999878934239e-05 ::
			9.99999999867220518e-05 ::
			9.99999999862202289e-05 ::
			9.99999999857962787e-05 ::
			9.99999999850009351e-05 ::
			9.99999999834475444e-05 ::
			9.99999999833636814e-05 :: []
		in
		if mx = 1.12458870341058947e-15 || mx = 3.95137481762342091e-16 then (
			ploop b p 0.0 4;
			Buffer.add_string b "00000000010000"
		) else if List.mem mx m9 then (
			ploop b (max (p -. mx) 0.0) 0.0 4;
			Buffer.add_string b "99999999990000"
		) else (
			ploop b p (0.1 ** 13.0) 14;
			Buffer.add_string b "0000"
		);
		Buffer.contents b
	end;;

module Parser = struct
	
	type float_type = Single | Date | Extended of (int64 * int) | Currency;;
	
	type token = ToEOF
		| ToSymbol of string
		| ToString of string
		| ToInteger of int64
		| ToFloat of float * float_type
		| ToWString of Unicode.utf16_string
		| ToChar of char
		| ToBinary of byte_array;;
	
	exception Error of int * string;;
	
	type t = {
		stream: char Stream.t;
		buffer: Buffer.t;
		mutable source_line: int;
		mutable token: token};;

	let error (self: t) (message: string) = (
		raise (Error (self.source_line, message))
	);;
	
	let skip_blanks (self: t): unit = (
		while
			match Stream.peek self.stream with
			| Some '\n' ->
				self.source_line <- self.source_line + 1;
				true
			| Some ('\x00'..'\t' | '\x0b'..' ') -> true
			| Some '!'..'\xff' -> false
			| None -> false
		do
			Stream.junk self.stream
		done
	);;
	
	let next_token (self: t): unit = (
		skip_blanks self;
		match Stream.peek self.stream with
		| Some ('A'..'Z' | 'a'..'z' | '_' as c) ->
			Stream.junk self.stream;
			Buffer.clear self.buffer;
			Buffer.add_char self.buffer c;
			while
				match Stream.peek self.stream with
				| Some ('A'..'Z' | 'a'..'z' | '_' | '0'..'9' as c) ->
					Stream.junk self.stream;
					Buffer.add_char self.buffer c;
					true
				| _ ->	false
			do () done;
			self.token <- ToSymbol (Buffer.contents self.buffer)
		| Some ('#' | '\'') ->
			let str = ref (Unicode.UTF16.create 16) in
			let len = ref 0 in
			let is_wide = ref false in
			let add x = (
				if x > 127 then is_wide := true;
				if !len = Unicode.UTF16.length !str then (
					str := Unicode.UTF16.append !str !str
				);
				Unicode.UTF16.set !str !len x;
				incr len;
			) in
			while
				match Stream.peek self.stream with
				| Some '#' ->
					Stream.junk self.stream;
					Buffer.clear self.buffer;
					while
						match Stream.peek self.stream with
						| Some ('0'..'9' as c) ->
							Stream.junk self.stream;
							Buffer.add_char self.buffer c;
							true
						| _ ->	false
					do () done;
					add (int_of_string (Buffer.contents self.buffer));
					true
				| Some '\'' ->
					Stream.junk self.stream;
					while
						match Stream.npeek 2 self.stream with
						| '\'' :: '\'' :: _ ->
							Stream.junk self.stream;
							Stream.junk self.stream;
							add (int_of_char '\'');
							true
						| '\'' :: _ ->
							Stream.junk self.stream;
							false
						| c :: _ ->
							Stream.junk self.stream;
							add (int_of_char c);
							true
						| [] ->
							false
					do () done;
					true
				| _ -> false
			do () done;
			if !is_wide then (
				let x = Unicode.UTF16.sub !str 0 !len in
				self.token <- ToWString x
			) else (
				let x = String.make !len '\x00' in
				for i = 0 to !len - 1 do
					x.[i] <- char_of_int (Unicode.UTF16.get !str i)
				done;
				self.token <- ToString x
			)
		| Some '$' ->
			Stream.junk self.stream;
			Buffer.clear self.buffer;
			Buffer.add_string self.buffer "0x";
			while
				match Stream.peek self.stream with
				| Some ('0'..'9' | 'A'..'F' | 'a'..'f' as c) ->
					Stream.junk self.stream;
					Buffer.add_char self.buffer c;
					true
				| _ ->	false
			do () done;
			let x = Int64.of_string (Buffer.contents self.buffer) in
			self.token <- ToInteger x
		| Some ('-' | '0'..'9' as c) ->
			Stream.junk self.stream;
			Buffer.clear self.buffer;
			Buffer.add_char self.buffer c;
			while
				match Stream.peek self.stream with
				| Some ('0'..'9' as c) ->
					Stream.junk self.stream;
					Buffer.add_char self.buffer c;
					true
				| _ ->	false
			do () done;
			let float_type c = (
				match c with
				| 'c' | 'C' -> Currency
				| 'd' | 'D' -> Date
				| 's' | 'S' -> Single
				| _ -> assert false
			) in
			begin match Stream.peek self.stream with
			| Some ('.' | 'e' | 'E' | '+' | '-' as c) ->
				Stream.junk self.stream;
				Buffer.add_char self.buffer c;
				while
					match Stream.peek self.stream with
					| Some ('0'..'9' | '.' | 'e' | 'E' | '+' | '-' as c) ->
						Stream.junk self.stream;
						Buffer.add_char self.buffer c;
						true
					| _ ->	false
				do () done;
				let s = Buffer.contents self.buffer in
				begin match Stream.peek self.stream with
				| Some ('c' | 'C' | 'd' | 'D' | 's' | 'S' as c) ->
					let x = float_of_string s in
					let t = float_type c in
					self.token <- ToFloat (x, t)
				| _ ->
					let x, y = !extended_of_string s in
					self.token <- ToFloat (x, Extended y)
				end
			| Some ('c' | 'C' | 'd' | 'D' | 's' | 'S' as c) ->
				let x = float_of_string (Buffer.contents self.buffer) in
				let t = float_type c in
				self.token <- ToFloat (x, t)
			| _ ->
				let x = Int64.of_string (Buffer.contents self.buffer) in
				self.token <- ToInteger x
			end
		| Some '{' ->
			Stream.junk self.stream;
			Buffer.clear self.buffer;
			let s = "0xXX" in
			let dat = ref (Bigarray.Array1.create
				Bigarray.int8_unsigned Bigarray.c_layout 4096)
			in
			let len = ref 0 in
			while
				skip_blanks self;
				match Stream.npeek 2 self.stream with
				| ('0'..'9' | 'A'..'F' | 'a'..'f' as hi) ::
					('0'..'9' | 'A'..'F' | 'a'..'f' as lo) :: _
				->
					Stream.junk self.stream;
					Stream.junk self.stream;
					s.[2] <- hi;
					s.[3] <- lo;
					let d = int_of_string s in
					Bigarray.Array1.set !dat !len d;
					incr len;
					true
				| '}' :: _ ->
					Stream.junk self.stream;
					false
				| _ -> false
			do () done;
			self.token <- ToBinary (Bigarray.Array1.sub !dat 0 !len)
		| Some c ->
			Stream.junk self.stream;
			self.token <- ToChar c
		| None ->
			self.token <- ToEOF
	);;

	let create (stream: char Stream.t): t = (
		let self = {
			stream = stream;
			buffer = Buffer.create 256;
			source_line = 1;
			token = ToEOF}
		in
		next_token self;
		self
	);;

	let source_line (self: t): int = self.source_line;;
	let token (self: t): token = self.token;;

	let token_symbol_is (self: t) (sym: string): bool = (
		match self.token with
		| ToSymbol s when String.uppercase s = sym -> true
		| _ -> false
	);;

	let token_char_is (self: t) (sym: char): bool = (
		match self.token with
		| ToChar c when c = sym -> true
		| _ -> false
	);;

end;;

let internal_binary_to_text
	?(name: string ref option)
	?(linefeed: string = "\n")
	?(indent: string = "  ")
	?(utf8: bool = false)
	(input: string -> int -> int -> int)
	(output: string -> unit)
	: unit =
(
	let nesting_level = ref 0 in
	let output_indent () = (
		for i = 1 to !nesting_level do
			output indent
		done
	) in
	let rec convert_header (is_root: bool): bool = (
		begin match Reader.read_component input with
		| Some (flags, child_pos, classname, objectname) ->
			output_indent ();
			output (
				if flags land (set_of_ff FfInherited) <> 0 then "inherited " else
				if flags land (set_of_ff FfInline) <> 0 then "inline " else
				"object ");
			if objectname <> "" then (
				output objectname;
				output ": "
			);
			output classname;
			if flags land (set_of_ff FfChildPos) <> 0 then (
				output " [";
				output (string_of_int child_pos);
				output "]"
			);
			output linefeed;
			begin match name with
			| Some v when is_root -> v := classname
			| _ -> ()
			end;
			true
		| None ->
			false
		end
	) and convert_value (in_list: bool): bool = (
		let line_length = 64 in
		let handle_string (x: Unicode.utf16_string): unit = (
			let length = Unicode.UTF16.length x in
			if length = 0 then output "\'\'" else (
				incr nesting_level;
				if length > line_length then (
					output linefeed;
					output_indent ()
				);
				let i = ref 0 in
				let k = ref 0 in
				let s = String.create 6 in
				while !i < length do
					let xi = Unicode.UTF16.get x !i in
					if xi >= 32 && xi <> int_of_char '\'' && (utf8 || xi <= 0x7f) then (
						let b = Buffer.create line_length in
						output "\'";
						while
							!i < length && !i - !k < line_length &&
							let xi = Unicode.UTF16.get x !i in
							xi >= 32 && xi <> int_of_char '\'' && (utf8 || xi <= 0x7f)
						do
							let code = Unicode.UTF16.get_code x i in
							let p = ref 0 in
							Unicode.UTF8.set_code s p code;
							Buffer.add_substring b s 0 !p;
						done;
						output (Buffer.contents b);
						output "\'"
					) else (
						output "#";
						output (string_of_int xi);
						incr i
					);
					if !i - !k >= line_length && (!i < length) then (
						output " +";
						output linefeed;
						output_indent ();
						k := !i
					);
				done;
				decr nesting_level
			)
		) in
		let handle_binary x = (
			let bytes_per_line = 32 in
			output "{";
			incr nesting_level;
			let hex = "0123456789ABCDEF" in
			let s = String.create 2 in
			let count = Bigarray.Array1.dim x in
			let multiline = count >= bytes_per_line in
			let blocks = (count + bytes_per_line - 1) / bytes_per_line in
			for i = 0 to blocks - 1 do
				if multiline then (
					output linefeed;
					output_indent ()
				);
				let len = (
					if i < blocks - 1 then bytes_per_line else
					count mod bytes_per_line
				) in
				let offset = i * bytes_per_line in
				for j = 0 to len - 1 do
					let d = Bigarray.Array1.get x (offset + j) in
					s.[0] <- hex.[d lsr 4];
					s.[1] <- hex.[d land 0x0f];
					output s
				done
			done;
			decr nesting_level;
			output "}"
		) in
		let handle_set () = (
			output "[";
			let rec loop first = (
				begin match Reader.read_set_element input with
				| Some x ->
					if not first then output ", ";
					begin match x with
					| `char x -> output "#"; output (string_of_int (int_of_char x))
					| `int x -> output (string_of_int x)
					| `tag x -> output x
					| `wchar x -> output "#"; output (string_of_int x)
					end;
					loop false
				| None -> ()
				end
			) in
			loop true;
			output "]"
		) in
		let handle_list () = (
			output "(";
			incr nesting_level;
			let rec loop () = (
				if convert_value true then loop ()
			) in
			loop ();
			decr nesting_level;
			output ")"
		) in
		let handle_collection () = (
			output "<";
			incr nesting_level;
			while
				match Reader.read_list input with
				| Some pos ->
					output linefeed;
					output_indent ();
					output "item";
					if pos >= 0 then (
						output " [";
						output (string_of_int pos);
						output "]"
					);
					output linefeed;
					incr nesting_level;
					while convert_property () do () done;
					decr nesting_level;
					output_indent ();
					output "end";
					true
				| None -> false
			do () done;
			decr nesting_level;
			output ">"
		) in
		let value = Reader.read_variant input in
		if in_list && value <> `null then (
			output linefeed;
			output_indent ()
		);
		begin match value with
		| `binary x -> handle_binary x
		| `bool x -> output (if x then "True" else "False")
		| `collection -> handle_collection ()
		| `char x -> handle_string (Unicode.UTF16.of_array [| int_of_char x |])
		| `currency x -> output (string_of_float (x *. 10000.0)); output "c"
		| `efloat x -> output (!string_of_extended x)
		| `float x -> output (string_of_float x)
		| `int x -> output (string_of_int x)
		| `int32 x -> output (Int32.to_string x)
		| `int64 x -> output (Int64.to_string x)
		| `list -> handle_list ()
		| `nil -> output "nil"
		| `null -> if not in_list then output "Null"
		| `set -> handle_set ()
		| `sfloat x -> output (string_of_float x); output "s"
		| `string x ->  handle_string (Unicode.utf16_of_utf8 x)
		| `tag x -> output x
		| `time x -> output (string_of_float x); output "d"
		| `wchar x -> handle_string (Unicode.UTF16.of_array [| x |])
		| `wstring x -> handle_string x
		end;
		value <> `null
	) and convert_property (): bool = (
		begin match Reader.read_property_name input with
		| Some property_name ->
			output_indent ();
			output property_name;
			output " = ";
			let (_: bool) = convert_value false in
			output linefeed;
			true
		| None ->
			false
		end
	) and convert_object (is_root: bool): bool = (
		if convert_header is_root then (
			incr nesting_level;
			while convert_property () do () done;
			while convert_object false do () done;
			decr nesting_level;
			output_indent ();
			output "end";
			output linefeed;
			true
		) else (
			if is_root then failwith "convert_object";
			false
		)
	) in
	Reader.read_signature input;
	let (_: bool) = convert_object true in ()
);;

let binary_to_text
	?(name: string ref option)
	?(linefeed: string option)
	?(indent: string option)
	?(utf8: bool option)
	(in_file: in_channel)
	(output: out_channel): unit =
(
	let input_f = input in_file in
	let output_f = output_string output in
	internal_binary_to_text ?name ?linefeed ?indent ?utf8 input_f output_f
);;

let resource_to_text
	?(resource_name: string ref option)
	?(name: string ref option)
	?(linefeed: string option)
	?(indent: string option)
	?(utf8: bool option)
	(in_file: in_channel)
	(output: out_channel)
	: unit =
(
	let input_f = input in_file in
	let output_f = output_string output in
	(* let output_f x = output_string output x; flush output in *)
	let resname = Reader.read_resource_header input_f in
	begin match resource_name with
	| Some v -> v := resname
	| None -> ()
	end;
	internal_binary_to_text ?name ?linefeed ?indent ?utf8 input_f output_f
);;

let internal_text_to_binary
	?(name: string ref option)
	?(no_utf16: bool = false)
	(input: char Stream.t)
	(output: string -> int -> int -> unit)
	: unit =
(
	let parser = Parser.create input in
	let rec convert_order_modifier (): int = (
		begin match Parser.token parser with
		| Parser.ToChar '[' ->
			Parser.next_token parser;
			begin match Parser.token parser with
			| Parser.ToInteger result ->
				Parser.next_token parser;
				begin match Parser.token parser with
				| Parser.ToChar ']' ->
					Parser.next_token parser;
					Int64.to_int result
				| _ -> Parser.error parser "']' expected"
				end
			| _ -> Parser.error parser "position expected"
			end
		| _ -> -1
		end
	) and convert_header (is_root: bool) (flags: filer_flags): unit = (
		let classname, objectname = (
			begin match Parser.token parser with
			| Parser.ToSymbol s ->
				Parser.next_token parser;
				begin match Parser.token parser with
				| Parser.ToChar ':' ->
					Parser.next_token parser;
					begin match Parser.token parser with
					| Parser.ToSymbol t ->
						Parser.next_token parser;
						t, s
					| _ -> Parser.error parser "class name expected"
					end
				| _ -> s, ""
				end
			| _ -> Parser.error parser "class/object name expected"
			end
		) in
		if is_root then (
			match name with
			| Some var -> var := classname
			| None -> ()
		);
		let position = convert_order_modifier () in
		let flags = (
			if position >= 0 then flags lor (set_of_ff FfChildPos) else flags)
		in
		Writer.write_component output flags position classname objectname
	) and convert_value (): unit = (
		begin match Parser.token parser with
		| Parser.ToString _ | Parser.ToWString _ ->
			let rec combine_wide_string s = (
				let s = (
					begin match Parser.token parser with
					| Parser.ToString u ->
						Unicode.UTF16.append s (Unicode.utf16_of_utf8 u)
					| Parser.ToWString w ->
						Unicode.UTF16.append s w
					| _ -> Parser.error parser "string expected"
					end
				) in
				Parser.next_token parser;
				begin match Parser.token parser with
				| Parser.ToChar '+' ->
					Parser.next_token parser;
					combine_wide_string s
				| _ -> s
				end
			) in
			let str = combine_wide_string Unicode.UTF16.empty in
			if no_utf16 then (
				Writer.write_string output (Unicode.utf8_of_utf16 str:> string)
			) else (
				Writer.write_wide_string output str
			)
		| Parser.ToSymbol s ->
			Parser.next_token parser;
			Writer.write_ident output s
		| Parser.ToInteger x ->
			Parser.next_token parser;
			Writer.write_int64 output x
		| Parser.ToFloat (x, t) ->
			Parser.next_token parser;
			begin match t with
			| Parser.Single -> Writer.write_single output x
			| Parser.Currency -> Writer.write_currency output (x /. 10000.0)
			| Parser.Date -> Writer.write_date output x
			| Parser.Extended e -> Writer.write_extended_image output e
			end
		| Parser.ToChar '[' ->
			Writer.write_set_begin output;
			Parser.next_token parser;
			let handle_string (s: Unicode.utf16_string) = (
				if Unicode.UTF16.length s <> 1 then (
					Parser.error parser "invalid character element"
				);
				let c = Unicode.UTF16.get s 0 in
				Writer.write_set_element output (`wchar c);
			) in
			let rec element (close: bool) = (
				begin match Parser.token parser with
				| Parser.ToChar ']' when close ->
					Parser.next_token parser
				| Parser.ToInteger x ->
					Parser.next_token parser;
					Writer.write_set_element output (`int (Int64.to_int x));
					comma_or_close ()
				| Parser.ToString x ->
					Parser.next_token parser;
					handle_string (Unicode.utf16_of_utf8 x);
					comma_or_close ()
				| Parser.ToWString x ->
					Parser.next_token parser;
					handle_string x;
					comma_or_close ()
				| Parser.ToSymbol x ->
					Parser.next_token parser;
					Writer.write_set_element output (`tag x);
					comma_or_close ()
				| _ -> Parser.error parser "element expected"
				end
			) and comma_or_close () = (
				begin match Parser.token parser with
				| Parser.ToChar ']' ->
					Parser.next_token parser
				| Parser.ToChar ',' ->
					Parser.next_token parser;
					element false
				| _ -> Parser.error parser "','/']' expected"
				end
			) in
			element true;
			Writer.write_set_end output
		| Parser.ToChar '(' -> (* list *)
			Parser.next_token parser;
			Writer.write_list_begin output;
			while not (Parser.token_char_is parser ')') do
				convert_value ()
			done;
			Parser.next_token parser;
			Writer.write_list_end output
		| Parser.ToBinary data -> (* binary *)
			Parser.next_token parser;
			Writer.write_binary output data
		| Parser.ToChar '<' -> (* collection *)
			Parser.next_token parser;
			Writer.write_collection_begin output;
			while not (Parser.token_char_is parser '>') do
				if Parser.token_symbol_is parser "ITEM" then (
					Parser.next_token parser;
					let order = convert_order_modifier () in
					if order >= 0 then (
						Writer.write_int output order
					);
					Writer.write_list_begin output;
					while not (Parser.token_symbol_is parser "END") do
						convert_property ()
					done;
					Parser.next_token parser;
					Writer.write_list_end output
				) else Parser.error parser "ITEM/'>' expected"
			done;
			Parser.next_token parser;
			Writer.write_collection_end output;
		| _ -> Parser.error parser "invalid value"
		end
	) and convert_property (): unit = (
		begin match Parser.token parser with
		| Parser.ToSymbol s ->
			Parser.next_token parser;
			let rec loop s = (
				if Parser.token_char_is parser '.' then (
					Parser.next_token parser;
					begin match Parser.token parser with
					| Parser.ToSymbol c ->
						Parser.next_token parser;
						loop (s ^ "." ^ c)
					| _ -> Parser.error parser "property name expected"
					end
				) else s
			) in
			let propname = loop s in
			Writer.write_property_name output propname;
			begin match Parser.token parser with
			| Parser.ToChar '=' ->
				Parser.next_token parser;
				convert_value ()
			| _ -> Parser.error parser "'=' expected"
			end
		| _ -> Parser.error parser "property name/END expected"
		end
	) and convert_object is_root: unit = (
		let flags: filer_flags = (
			if Parser.token_symbol_is parser "INHERITED" then (
				set_of_ff FfInherited
			) else if Parser.token_symbol_is parser "INLINE" then (
				set_of_ff FfInline
			) else if Parser.token_symbol_is parser "OBJECT" then (
				0
			) else (
				Parser.error parser "OBJECT expected"
			)
		) in
		Parser.next_token parser;
		convert_header is_root flags;
		while not (Parser.token_symbol_is parser "END") &&
			not (Parser.token_symbol_is parser "OBJECT") &&
			not (Parser.token_symbol_is parser "INHERITED") &&
			not (Parser.token_symbol_is parser "INLINE")
		do
			convert_property ()
		done;
		Writer.write_list_end output;
		while not (Parser.token_symbol_is parser "END") do
			convert_object false;
			Parser.next_token parser; (* END of child component *)
		done;
		Writer.write_list_end output
	) in
	Writer.write_signature output;
	convert_object true
);;

let text_to_binary
	?(name: string ref option)
	?(no_utf16: bool option)
	(input: in_channel)
	(out_file: out_channel)
	: unit =
(
	let input_stream = Stream.of_channel input in
	let output_f = output out_file in
	internal_text_to_binary ?name ?no_utf16 input_stream output_f
);;

let text_to_resource
	?(resource_name: string option)
	?(name: string ref option)
	?(no_utf16: bool option)
	(input: in_channel)
	(out_file: out_channel)
	: unit =
(
	let buffer = Buffer.create 4096 in
	let classname = ref "" in
	internal_text_to_binary ~name:classname ?no_utf16
		(Stream.of_channel input) (Buffer.add_substring buffer);
	begin match name with
	| Some v -> v := !classname
	| None -> ()
	end;
	let size = Buffer.length buffer in
	let resource_name =
		begin match resource_name with
		| Some v -> v
		| None -> String.uppercase !classname
		end
	in
	Writer.write_resource_header (output out_file) resource_name size;
	Buffer.output_buffer out_file buffer
);;
