type 'v writer = [`writing of (unit -> unit) list ref *
	(string -> 'v -> unit)];;
type 'v reader = [`reading of string ref * 'v ref *
	(unit -> unit) * (float -> float)];;
type 'v serializer = ['v writer | 'v reader];;

(* primitives *)

let write (w: 'a writer) ?(key: string = "") (value: 'a): unit = (
	let `writing (_, write) = w in
	write key value
);;

let peek (r: 'a reader): string * 'a = (
	let `reading (next_key, next_value, _, _) = r in
	!next_key, !next_value
);;

let junk (r: 'a reader): unit = (
	let `reading (_, _, junk, _) = r in
	junk ()
);;

let rec junk_structure (r: 'a reader): unit = (
	let `reading (_, next_value, junk, _) = r in
	match !next_value with
	| `record | `list _ ->
		junk ();
		while
			match !next_value with
			| `end_of_record | `end_of_list -> true
			| _ -> false
		do
			junk_structure r
		done;
		junk ()
	| _ -> junk ()
);;

(* structured *)

let make_io variant_of of_variant ?(key = "") value (s: 'a serializer) = (
	match s with
	| `writing (es, f) ->
		let e () = f key (variant_of value) in
		if key = "" then e () else es := e :: !es;
		value
	| `reading (k, v, _, _) ->
		if !k = key then of_variant !v else value
);;

let variant_of_bool x = (
	`bool x
);;

let bool_of_variant x = (
	match x with
	| `bool v -> v
	| _ -> failwith "bool_of_variant"
);;

let io_bool ?key v s = make_io variant_of_bool bool_of_variant ?key v s;;

let variant_of_char x = (
	`char x
);;

let char_of_variant x = (
	match x with
	| `char v -> v
	| _ -> failwith "char_of_variant"
);;

let io_char ?key v s = make_io variant_of_char char_of_variant ?key v s;;

let variant_of_float x = (
	`float x
);;

let float_of_variant x = (
	match x with
	| `float v -> v
	| `int v -> float_of_int v
	| _ -> failwith "float_of_variant"
);;

let io_float ?key v s = make_io variant_of_float float_of_variant ?key v s;;

let variant_of_int x = (
	`int x
);;

let int_of_variant x = (
	match x with
	| `int v -> v
	| _ -> failwith "int_of_variant"
);;

let io_int ?key v s = make_io variant_of_int int_of_variant ?key v s;;

let variant_of_string x = (
	`string x
);;

let string_of_variant x = (
	match x with
	| `char v -> String.make 1 v
	| `string v -> v
	| _ -> failwith "string_of_variant"
);;

let io_string ?key v s = make_io variant_of_string string_of_variant ?key v s;;

let variant_of_tag x = (
	`tag x
);;

let tag_of_variant x = (
	match x with
	| `tag v -> v
	| _ -> failwith "tag_of_variant"
);;

let io_tag ?key v s = make_io variant_of_tag tag_of_variant ?key v s;;

let io_time ?(key = "") value (s: 'a serializer) = (
	match s with
	| `writing (es, f) ->
		let e () = f key (`time value) in
		if key = "" then e () else es := e :: !es;
		value
	| `reading (k, v, _, time_of_float) ->
		if !k = key then (
			match !v with
			| `time x -> x
			| `float x -> time_of_float x
			| _ -> failwith "io_time"
		) else value
);;

let io_option (io_element: ?key:string -> 'b -> 'v serializer -> 'b) (proto: 'b)
	?(key = "") (value: 'b option) (s: 'a serializer): 'b option =
(
	match s with
	| `writing (es, f) ->
		let e () = (
			begin match value with
			| Some value ->
				f key `some;
				let (_: 'b) = io_element ?key:None value s in ()
			| None ->
				f key `none
			end
		) in
		if key = "" then e () else es := e :: !es;
		value
	| `reading (k, _, _, _) ->
		if !k = key then Some (io_element ~key proto s) else value
);;

let io_record (iter: 'a -> 'v serializer -> 'a)
	?(key: string = "") value (serializer: 'v serializer) =
(
	match serializer with
	| `writing (es, f) ->
		let e () = (
			f key `record;
			es := [];
			iter value serializer;
			List.iter (fun e -> e ()) !es;
			f "" `end_of_record
		) in
		if key = "" then e () else es := e :: !es;
		value
	| `reading (next_key, next_value, junk, _) as r ->
		if !next_key <> key then value else
		match !next_value with
		| `record ->
			junk ();
			let rec loop v = (
				if !next_value = `end_of_record then v else (
					let v2 = iter v serializer in
					junk_structure r;
					loop v2
				)
			) in
			loop value
		| `end_of_record -> value
		| _ -> failwith "io_record"	
);;

let check_is_record
	(io_element: ?key:string -> 'b -> 'v serializer -> 'b) (proto: 'b): bool =
(
	let module Local = struct exception Result of bool end in
	let w: [> 'v writer] =
		`writing (
			ref [],
			begin fun (_: string) (value: 'v) ->
				match value with
				| `record -> raise (Local.Result true)
				| _ -> raise (Local.Result false)
			end
		)
	in
	try io_element proto w; false with
	| Local.Result result -> result
)

let io_list (io_element: ?key:string -> 'b -> 'v serializer -> 'b) (proto: 'b)
	?(key: string = "") (value: 'b list) (serializer: 'v serializer) =
(
	match serializer with
	| `writing (es, f) ->
		let e () = (
			if value = [] then (
				let is_record = check_is_record io_element proto in
				f key (`list (if is_record then `record else `any))
			) else (
				f key (`list `any);
				let rec loop xs = (
					match xs with
					| [] -> ()
					| x :: xr -> let (_: 'b) = io_element x serializer in loop xr
				) in
				loop value
			);
			f "" `end_of_list
		) in
		if key = "" then e () else es := e :: !es;
		value
	| `reading (next_key, next_value, junk, _) as r ->
		if !next_key <> key then value else
		match !next_value with
		| `list _ ->
			junk ();
			let rec loop rs = (
				if !next_value = `end_of_list then List.rev rs else (
					let rs2 = io_element proto serializer :: rs in
					junk_structure r;
					loop rs2
				)
			) in
			loop []
		| `end_of_list -> value
		| _ -> failwith "io_list"
);;
