#include <yaml.h>
#include <stdint.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/alloc.h>
#include <caml/fail.h>

static CAMLprim void raise_yaml_error(yaml_error_type_t t, char const *s)
{
	CAMLparam0();
	CAMLlocal3(arg1, arg2, exn);
	value *exception = caml_named_value("YamlError");
	arg1 = Val_int(t);
	arg2 = caml_copy_string(s);
#if 0
	value args[2];
	args[0] = arg1;
	args[1] = arg2;
	caml_raise_with_args(*exception, 2, args);
#else
	/* O'Caml 3.10.1 (latest of G'Caml) does not have caml_raise_with_args. */
	exn = caml_alloc(3, 0);
	Field(exn, 0) = *exception;
	Field(exn, 1) = arg1;
	Field(exn, 2) = arg2;
	caml_raise(exn);
#endif
	CAMLreturn0;
}

static int mlyaml_x_compare(value left, value right)
{
	CAMLparam2(left, right);
	intptr_t la = (intptr_t)(Data_custom_val(left));
	intptr_t ra = (intptr_t)(Data_custom_val(right));
	int result = (la < ra)? -1 : (la > ra)? 1 : 0;
	CAMLreturn(result);
}

static long mlyaml_x_hash(value data)
{
	CAMLparam1(data);
	long result = (intptr_t)(Data_custom_val(data));
	CAMLreturn(result);
}

static CAMLprim value Val_mark(yaml_mark_t *x)
{
	CAMLparam0();
	CAMLlocal1(result);
	result = caml_alloc_tuple(3);
	Field(result, 0) = Val_int(x->index);
	Field(result, 1) = Val_int(x->line);
	Field(result, 2) = Val_int(x->column);
	CAMLreturn(result);	
}

/* version */

CAMLprim value mlyaml_get_version_string()
{
	CAMLparam0();
	CAMLlocal1(result);
	char const *ver = yaml_get_version_string();
	result = caml_copy_string(ver);
	CAMLreturn(result);
}

CAMLprim value mlyaml_get_version()
{
	CAMLparam0();
	CAMLlocal1(result);
	int major, minor, patch;
	yaml_get_version(&major, &minor, &patch);
	result = caml_alloc_tuple(3);
	Field(result, 0) = Val_int(major);
	Field(result, 1) = Val_int(minor);
	Field(result, 2) = Val_int(patch);
	CAMLreturn(result);
}

/* parser */

static inline yaml_parser_t *parser_val(value data)
{
	return (yaml_parser_t *)(Data_custom_val(data));
}

static void mlyaml_parser_finalize(value r);

static struct custom_operations parser_ops = {
	.identifier = "jp.halfmoon.panathenaia.yaml",
	.finalize = mlyaml_parser_finalize,
	.compare = mlyaml_x_compare,
	.hash = mlyaml_x_hash,
	.serialize = custom_serialize_default,
	.deserialize = custom_deserialize_default};

static void mlyaml_parser_finalize(value data)
{
	CAMLparam1(data);
	yaml_parser_t *e = parser_val(data);
	caml_remove_global_root((value *)&e->read_handler_data);
	yaml_parser_delete(e);
	CAMLreturn0;
}

static int yaml_read_handler(void *data,
	unsigned char *buffer, size_t size, size_t *length)
{
	CAMLparam0();
	CAMLlocal2(handler, s);
	handler = (value)data; /* data will be broken */
	s = caml_alloc_string(size); /* may invoke gc */
	*length = Int_val(caml_callback3(handler,
		s, Val_int(0), Val_int(size)));
	memcpy(buffer, String_val(s), *length);
	CAMLreturn(1);
}

CAMLprim value mlyaml_parser_initialize(value read)
{
	CAMLparam1(read);
	CAMLlocal1(result);
	result = alloc_custom(&parser_ops, sizeof(yaml_parser_t), 0, 1);
	yaml_parser_t *p = parser_val(result);
	if(yaml_parser_initialize(p) == 0){
		raise_yaml_error(p->error, p->problem);
	}
	yaml_parser_set_input(p, yaml_read_handler, (void *)read);
	caml_register_global_root((value *)&p->read_handler_data);
	CAMLreturn(result);
}

CAMLprim value mlyaml_parser_set_encoding(value parser, value encoding)
{
	CAMLparam2(parser, encoding);
	yaml_parser_t *p = parser_val(parser);
	yaml_parser_set_encoding(p, Int_val(encoding));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_parser_parse(value parser)
{
	enum {
		result, ev, data,
		version_directive, version_directive_option,
		tag_directive, tag_directives, cons,
		local_count
	};
	CAMLparam1(parser);
	CAMLlocalN(x, local_count);
	yaml_event_t event;
	yaml_parser_t *p = parser_val(parser);
	if(yaml_parser_parse(p, &event) == 0){
		p = parser_val(parser); /* if gc was invoked on callback handler */
		raise_yaml_error(p->error, p->problem);
	}
	switch(event.type){
	case YAML_NO_EVENT:
		x[ev] = Val_int(0);
		break;
	case YAML_STREAM_START_EVENT:
		x[ev] = caml_alloc(1, 0);
		Field(x[ev], 0) = Val_int(event.data.stream_start.encoding);
		break;
	case YAML_STREAM_END_EVENT:
		x[ev] = Val_int(1);
		break;
	case YAML_DOCUMENT_START_EVENT:
		if(event.data.document_start.version_directive == NULL){
			x[version_directive_option] = Val_int(0); /* None */
		}else{
			x[version_directive] = caml_alloc_tuple(2);
			Field(x[version_directive], 0) =
				event.data.document_start.version_directive->major;
			Field(x[version_directive], 1) =
				event.data.document_start.version_directive->minor;
			x[version_directive_option] = caml_alloc(1, 0); /* Some */
			Field(x[version_directive_option], 0) = x[version_directive];
		}
		x[tag_directives] = Val_emptylist;
		if(event.data.document_start.tag_directives.start <
			event.data.document_start.tag_directives.end)
		{
			yaml_tag_directive_t *tdi;
			tdi = event.data.document_start.tag_directives.start;
			while(tdi != event.data.document_start.tag_directives.end){
				x[tag_directive] = caml_alloc_tuple(2);
				Field(x[tag_directive], 0) =
					caml_copy_string((char *)tdi->handle);
				Field(x[tag_directive], 1) =
					caml_copy_string((char *)tdi->prefix);
				x[cons] = caml_alloc(2, Tag_cons);
				Field(x[cons], 0) = x[tag_directive];
				Field(x[cons], 1) = x[tag_directives];
				x[tag_directives] = x[cons];
				++ tdi;
			}
		}
		x[ev] = caml_alloc(3, 1);
		Field(x[ev], 0) = x[version_directive_option];
		Field(x[ev], 1) = x[tag_directives];
		Field(x[ev], 2) = Val_int(event.data.document_start.implicit);
		break;
	case YAML_DOCUMENT_END_EVENT:
		x[ev] = caml_alloc(1, 2);
		Field(x[ev], 0) = Val_int(event.data.document_end.implicit);
		break;
	case YAML_ALIAS_EVENT:
		x[ev] = caml_alloc(1, 3);
		Field(x[ev], 0) = caml_copy_string((char *)event.data.alias.anchor);
		break;
	case YAML_SCALAR_EVENT:
		x[data] = caml_alloc_string(event.data.scalar.length);
		memcpy(String_val(x[data]), event.data.scalar.value,
			event.data.scalar.length);
		x[ev] = caml_alloc(6, 4);
		Field(x[ev], 0) = caml_copy_string(
			(event.data.scalar.anchor != NULL) ?
			(char *)event.data.scalar.anchor : "");
		Field(x[ev], 1) = caml_copy_string(
			(event.data.scalar.tag != NULL) ?
			(char *)event.data.scalar.tag : "");
		Field(x[ev], 2) = x[data];
		Field(x[ev], 3) = Val_int(event.data.scalar.plain_implicit);
		Field(x[ev], 4) = Val_int(event.data.scalar.quoted_implicit);
		Field(x[ev], 5) = Val_int(event.data.scalar.style);
		break;
	case YAML_SEQUENCE_START_EVENT:
		x[ev] = caml_alloc(4, 5);
		Field(x[ev], 0) = caml_copy_string(
			(event.data.sequence_start.anchor != NULL) ?
			(char *)event.data.sequence_start.anchor : "");
		Field(x[ev], 1) = caml_copy_string(
			(event.data.sequence_start.tag != NULL) ?
			(char *)event.data.sequence_start.tag : "");
		Field(x[ev], 2) = Val_int(event.data.sequence_start.implicit);
		Field(x[ev], 3) = Val_int(event.data.sequence_start.style);
		break;
	case YAML_SEQUENCE_END_EVENT:
		x[ev] = Val_int(2);
		break;
	case YAML_MAPPING_START_EVENT:
		x[ev] = caml_alloc(4, 6);
		Field(x[ev], 0) = caml_copy_string(
			(event.data.mapping_start.anchor != NULL) ?
			(char *)event.data.mapping_start.anchor : "");
		Field(x[ev], 1) = caml_copy_string(
			(event.data.mapping_start.tag != NULL) ?
			(char *)event.data.mapping_start.tag : "");
		Field(x[ev], 2) = Val_int(event.data.mapping_start.implicit);
		Field(x[ev], 3) = Val_int(event.data.mapping_start.style);
		break;
	case YAML_MAPPING_END_EVENT:
		x[ev] = Val_int(3);
		break;
	default:
		failwith("mlyaml_parser_parse: unknown event");
	}
	x[result] = caml_alloc_tuple(3);
	Field(x[result], 0) = x[ev];
	Field(x[result], 1) = Val_mark(&event.start_mark);
	Field(x[result], 2) = Val_mark(&event.end_mark);
	yaml_event_delete(&event);
	CAMLreturn(x[result]);
}

CAMLprim value mlyaml_parser_problem_mark(value parser)
{
	CAMLparam1(parser);
	CAMLlocal1(result);
	yaml_parser_t *p = parser_val(parser);
	result = Val_mark(&p->problem_mark);
	CAMLreturn(result);
}

/* emitter */

static inline yaml_emitter_t *emitter_val(value data)
{
	return (yaml_emitter_t *)(Data_custom_val(data));
}

static void mlyaml_emitter_finalize(value r);

static struct custom_operations emitter_ops = {
	.identifier = "jp.halfmoon.panathenaia.yaml",
	.finalize = mlyaml_emitter_finalize,
	.compare = mlyaml_x_compare,
	.hash = mlyaml_x_hash,
	.serialize = custom_serialize_default,
	.deserialize = custom_deserialize_default};

static void mlyaml_emitter_finalize(value data)
{
	CAMLparam1(data);
	yaml_emitter_t *e = emitter_val(data);
	caml_remove_global_root((value *)&e->write_handler_data);
	yaml_emitter_delete(e);
	CAMLreturn0;
}

static int yaml_write_handler(void *data, unsigned char *buffer, size_t size)
{
	CAMLparam0();
	CAMLlocal2(handler, s);
	handler = (value)data;
	s = caml_alloc_string(size);
	memcpy(String_val(s), buffer, size);
	caml_callback3(handler, s, Val_int(0), Val_int(size));
	CAMLreturn(1);
}

CAMLprim value mlyaml_emitter_initialize(value write)
{
	CAMLparam1(write);
	CAMLlocal1(result);
	result = alloc_custom(&emitter_ops, sizeof(yaml_emitter_t), 0, 1);
	yaml_emitter_t *e = emitter_val(result);
	if(yaml_emitter_initialize(e) == 0){
		raise_yaml_error(e->error, e->problem);
	}
	yaml_emitter_set_output(e, yaml_write_handler, (void *)write);
	caml_register_global_root((value *)&e->write_handler_data);
	CAMLreturn(result);
}

CAMLprim value mlyaml_emitter_set_break(value emitter, value the_break)
{
	CAMLparam2(emitter, the_break);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_emitter_set_break(e, Int_val(the_break));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_set_canonical(value emitter, value canonical)
{
	CAMLparam2(emitter, canonical);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_emitter_set_canonical(e, Int_val(canonical));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_set_indent(value emitter, value indent)
{
	CAMLparam2(emitter, indent);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_emitter_set_indent(e, Int_val(indent));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_set_unicode(value emitter, value unicode)
{
	CAMLparam2(emitter, unicode);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_emitter_set_unicode(e, Int_val(unicode));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_set_width(value emitter, value width)
{
	CAMLparam2(emitter, width);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_emitter_set_width(e, Int_val(width));
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_emit(value emitter, value event)
{
	CAMLparam2(emitter, event);
	yaml_emitter_t *e = emitter_val(emitter);
	yaml_event_t ev;
	if(Is_long(event) && Int_val(event) == 0){
		/* YAML_NO_EVENT */
		failwith("mlyaml_emitter_emit: YAML_NO_EVENT");
	} else if(Is_block(event) && Tag_val(event) == 0){
		/* YAML_STREAM_START_EVENT */
		yaml_encoding_t encoding = Int_val(Field(event, 0));
		yaml_stream_start_event_initialize(&ev, encoding);
	} else if(Is_long(event) && Int_val(event) == 1){
		/* YAML_STREAM_END_EVENT */
		yaml_stream_end_event_initialize(&ev);
	} else if(Is_block(event) && Tag_val(event) == 1){
		/* YAML_DOCUMENT_START_EVENT */
		yaml_version_directive_t version_directive_body;
		yaml_version_directive_t *version_directive;
		yaml_tag_directive_t *tag_directives_start;
		yaml_tag_directive_t *tag_directives_end;
		size_t tdc;
		value tdi;
		value vd = Field(event, 0);
		value tag_directives = Field(event, 1);
		int implicit = Int_val(Field(event, 2));
		if(Is_long(vd)){
			/* None */
			version_directive = NULL;
		}else{
			/* Some (major, minor) */
			value v = Field(vd, 0);
			version_directive_body.major = Int_val(Field(v, 0));
			version_directive_body.minor = Int_val(Field(v, 1));
			version_directive = &version_directive_body;
		}
		tdi = tag_directives;
		tdc = 0;
		while(Is_block(tdi)){
			++ tdc;
			tdi = Field(tdi, 1);
		}
		if(tdc == 0){
			tag_directives_start = NULL;
			tag_directives_end = NULL;
		}else{
			yaml_tag_directive_t *i;
			tag_directives_start = malloc(tdc * sizeof(yaml_tag_directive_t));
			tag_directives_end = tag_directives_start + tdc;
			tdi = tag_directives;
			i = tag_directives_start;
			while(Is_block(tdi)){
				value item = Field(tdi, 0);
				i->handle = (yaml_char_t *)String_val(Field(item, 0));
				i->prefix = (yaml_char_t *)String_val(Field(item, 1));
				++ i;
				tdi = Field(tdi, 1);
			}
		}
		yaml_document_start_event_initialize(&ev, version_directive,
			tag_directives_start, tag_directives_end, implicit);
		free(tag_directives_start);
	} else if(Is_block(event) && Tag_val(event) == 2){
		/* YAML_DOCUMENT_END_EVENT */
		int implicit = Int_val(Field(event, 0));
		yaml_document_end_event_initialize(&ev, implicit);
	} else if(Is_block(event) && Tag_val(event) == 3){
		/* YAML_ALIAS_EVENT */
		yaml_char_t *anchor = (yaml_char_t *)String_val(Field(event, 0));
		yaml_alias_event_initialize(&ev, anchor);
	} else if(Is_block(event) && Tag_val(event) == 4){
		/* YAML_SCALAR_EVENT */
		yaml_char_t *anchor = (yaml_char_t *)String_val(Field(event, 0));
		yaml_char_t *tag = (yaml_char_t *)String_val(Field(event, 1));
		yaml_char_t *data = (yaml_char_t *)String_val(Field(event, 2));
		int length = caml_string_length(Field(event, 2));
		int plain_implicit = Int_val(Field(event, 3));
		int quoted_implicit = Int_val(Field(event, 4));
		yaml_scalar_style_t style = Int_val(Field(event, 5));
		if(anchor != NULL && *anchor == '\0') anchor = NULL;
		if(tag != NULL && *tag == '\0') tag = NULL;
		yaml_scalar_event_initialize(&ev, anchor, tag,
			data, length, plain_implicit, quoted_implicit, style);
	} else if(Is_block(event) && Tag_val(event) == 5){
		/* YAML_SEQUENCE_START_EVENT */
		yaml_char_t *anchor = (yaml_char_t *)String_val(Field(event, 0));
		yaml_char_t *tag = (yaml_char_t *)String_val(Field(event, 1));
		int implicit = Int_val(Field(event, 2));
		yaml_sequence_style_t style = Int_val(Field(event, 3));
		if(anchor != NULL && *anchor == '\0') anchor = NULL;
		if(tag != NULL && *tag == '\0') tag = NULL;
		yaml_sequence_start_event_initialize(&ev, anchor, tag, implicit, style);
	} else if(Is_long(event) && Int_val(event) == 2){
		/* YAML_SEQUENCE_END_EVENT */
		yaml_sequence_end_event_initialize(&ev);
	} else if(Is_block(event) && Tag_val(event) == 6){
		/* YAML_MAPPING_START_EVENT */
		yaml_char_t *anchor = (yaml_char_t *)String_val(Field(event, 0));
		yaml_char_t *tag = (yaml_char_t *)String_val(Field(event, 1));
		int implicit = Int_val(Field(event, 2));
		yaml_mapping_style_t style = Int_val(Field(event, 3));
		if(anchor != NULL && *anchor == '\0') anchor = NULL;
		if(tag != NULL && *tag == '\0') tag = NULL;
		yaml_mapping_start_event_initialize(&ev, anchor, tag, implicit, style);
	} else if(Is_long(event) && Int_val(event) == 3){
		/* YAML_MAPPING_END_EVENT */
		yaml_mapping_end_event_initialize(&ev);
	} else {
		failwith("mlyaml_emitter_emit: unknown event");
	}
	if(yaml_emitter_emit(e, &ev) == 0){
		e = emitter_val(emitter); /* if gc was invoked on callback handler */
		raise_yaml_error(e->error, e->problem);
	}
	CAMLreturn(Val_unit);
}

CAMLprim value mlyaml_emitter_flush(value emitter)
{
	CAMLparam1(emitter);
	yaml_emitter_t *e = emitter_val(emitter);
	if(yaml_emitter_flush(e) == 0){
		raise_yaml_error(e->error, e->problem);
	}
	CAMLreturn(Val_unit);
}
