open Yaml;;

module Writer = struct

	let create (typename: string) (write: yaml_write_handler)
		: [> `writing of (unit -> unit) list ref * (string -> 'v -> unit)] =
	(
		let emitter = yaml_emitter_initialize write in
		yaml_emitter_set_unicode emitter true;
		yaml_emitter_emit emitter (YAML_STREAM_START_EVENT
			YAML_UTF8_ENCODING);
		yaml_emitter_emit emitter (YAML_DOCUMENT_START_EVENT
			(None, [], typename = ""));
		let close () = (
			yaml_emitter_emit emitter (YAML_DOCUMENT_END_EVENT true);
			yaml_emitter_emit emitter YAML_STREAM_END_EVENT;
			yaml_emitter_flush emitter;
		) in
		let state = ref [] in
		`writing (ref [], (fun key value ->
			if key <> "" then (
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT
					("", "", key, true, true, YAML_ANY_SCALAR_STYLE))
			);
			let tag = (
				if !state = [] && typename <> "" then typename else ""
			) in
			begin match value with
			| `record ->
				yaml_emitter_emit emitter (YAML_MAPPING_START_EVENT
					("", tag, tag = "", YAML_ANY_MAPPING_STYLE));
				state := `record :: !state
			| `end_of_record ->
				yaml_emitter_emit emitter YAML_MAPPING_END_EVENT;
				begin match !state with
				| `record :: sr ->
					state := sr;
					if sr = [] then close ()
				| _ -> failwith "Writer.create:end_of_record"
				end
			| `list _ ->
				yaml_emitter_emit emitter (YAML_SEQUENCE_START_EVENT
					("", tag, tag = "", YAML_ANY_SEQUENCE_STYLE));
				state := `list :: !state
			| `end_of_list ->
				yaml_emitter_emit emitter YAML_SEQUENCE_END_EVENT;
				begin match !state with
				| `list :: sr ->
					state := sr;
					if sr = [] then close ()
				| _ -> failwith "Writer.create:end_of_list"
				end
			| `some ->
				()
			| `none ->
				if !state = [] then close ()
			| `char x ->
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT (
					"", tag, String.make 1 x, tag = "", tag = "",
					YAML_SINGLE_QUOTED_SCALAR_STYLE));
				if !state = [] then close ()
			| `float x ->
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT (
					"", tag, string_of_float x, tag = "", tag = "",
					YAML_ANY_SCALAR_STYLE));
				if !state = [] then close ()
			| `int x ->
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT (
					"", tag, string_of_int x, tag = "", tag = "",
					YAML_ANY_SCALAR_STYLE));
				if !state = [] then close ()
			| `string x ->
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT (
					"", tag, x, tag = "", tag = "", (
					if tag = "" then YAML_DOUBLE_QUOTED_SCALAR_STYLE else
					YAML_LITERAL_SCALAR_STYLE)));
				if !state = [] then close ()
			| `tag x ->
				yaml_emitter_emit emitter (YAML_SCALAR_EVENT (
					"", tag, x, tag = "", tag = "",
					YAML_PLAIN_SCALAR_STYLE));
				if !state = [] then close ()
			end
		))
	);;

end;;

module Reader = struct

	let create (typename: string) (read: yaml_read_handler)
		: [> `reading of string ref * 'v ref * (unit -> unit) *
			(float -> float)] =
	(
		let parser = yaml_parser_initialize read in
		let state = ref [] in
		let next_key = ref "" in
		let next_value = ref `end_of_record in
		let close () = (
			let event, _, _ = yaml_parser_parse parser in
			begin match event with
			| YAML_DOCUMENT_END_EVENT _ -> ()
			| _ -> failwith "Reader.create:expected YAML_DOCUMENT_END_EVENT"
			end;
			let event, _, _ = yaml_parser_parse parser in
			begin match event with
			| YAML_STREAM_END_EVENT -> ()
			| _ -> failwith "Reader.create:expected YAML_STREAM_END_EVENT"
			end
		) in
		let check_type tag = (
			if tag <> typename then (
				failwith ("Reader.create: mismatch \'" ^ tag ^ "\'")
			)
		) in
		let rec advance () = (
			let event, _, _ = yaml_parser_parse parser in
			begin match event with
			| YAML_NO_EVENT ->
				failwith "Reader.create:YAML_NO_EVENT"
			| YAML_STREAM_START_EVENT _ ->
				advance ()
			| YAML_STREAM_END_EVENT ->
				failwith "Reader.create:YAML_STREAM_END_EVENT"
			| YAML_DOCUMENT_START_EVENT _ ->
				advance ()
			| YAML_DOCUMENT_END_EVENT _ ->
				failwith "Reader.create:YAML_DOCUMENT_END_EVENT"
			| YAML_ALIAS_EVENT _ ->
				failwith "Reader.create:YAML_ALIAS_EVENT"
			| YAML_SCALAR_EVENT (_, tag, value, _, _, style) ->
				let convert value style = (
					begin match style with
					| YAML_PLAIN_SCALAR_STYLE
					| YAML_ANY_SCALAR_STYLE ->
						(
							try `int (int_of_string value) with Failure _ ->
							try `float (float_of_string value) with Failure _ ->
							`tag value
						)
					| YAML_SINGLE_QUOTED_SCALAR_STYLE
					| YAML_DOUBLE_QUOTED_SCALAR_STYLE
					| YAML_LITERAL_SCALAR_STYLE
					| YAML_FOLDED_SCALAR_STYLE ->
						if String.length value = 1 then `char value.[0] else
						`string value
					end
				) in
				begin match !state with
				| `record1 :: sr ->
					next_key := value;
					state := `record2 :: sr;
					advance ()
				| `record2 :: sr ->
					next_value := convert value style;
					state := `record1 :: sr;
				| `list :: _ ->
					next_key := "";
					next_value := convert value style
				| [] ->
					check_type tag;
					next_key := "";
					next_value := convert value style;
					close ()
				end
			| YAML_SEQUENCE_START_EVENT (_, tag, _, _) ->
				if !state = [] then check_type tag;
				next_value := `list `any;
				state := `list :: !state
			| YAML_SEQUENCE_END_EVENT ->
				next_key := "";
				next_value := `end_of_list;
				begin match !state with
				| `list :: `record2 :: sr ->
					state := `record1 :: sr
				| `list :: sr ->
					state := sr;
					if sr = [] then close ()
				| _ -> failwith "Reader.create:YAML_SEQUENCE_END_EVENT"
				end
			| YAML_MAPPING_START_EVENT (_, tag, _, _) ->
				if !state = [] then check_type tag;
				next_value := `record;
				state := `record1 :: !state
			| YAML_MAPPING_END_EVENT ->
				next_key := "";
				next_value := `end_of_record;
				begin match !state with
				| `record1 :: `record2 :: sr ->
					state := `record1 :: sr
				| `record1 :: sr ->
					state := sr;
					if sr = [] then close ()
				| _ -> failwith "Reader.create:YAML_MAPPING_END_EVENT"
				end
			end
		) in
		advance ();
		`reading (next_key, next_value, advance, (fun x -> x))
	);;

end;;
